from PyPDF2 import PdfReader

class Paper(object):

    def __init__(self, pdf_obj: PdfReader) -> None:
        self._pdf_obj = pdf_obj
        self._paper_meta = self._pdf_obj.metadata

    def iter_pages(self, iter_text_len: int = 3000):
        page_idx = 0
        for page in self._pdf_obj.pages:
            txt = page.extract_text()
            for i in range((len(txt) // iter_text_len) + 1):
                yield page_idx, i, txt[i * iter_text_len:(i + 1) * iter_text_len]
            page_idx += 1




from PyPDF2 import PdfReader
import openai
import nonebot
from .prompt import BASE_POINTS, READING_PROMT_V2
from .paper import Paper
from .model_interface import OpenAIModel


# Setting the API key to use the OpenAI API
class PaperReader:
    """
    A class for summarizing research papers using the OpenAI API.

    Attributes:
        openai_key (str): The API key to use the OpenAI API.
        token_length (int): The length of text to send to the API at a time.
        model (str): The GPT model to use for summarization.
        points_to_focus (str): The key points to focus on while summarizing.
        verbose (bool): A flag to enable/disable verbose logging.

    """

    def __init__(self, openai_key, token_length=4000, model="gpt-3.5-turbo",
                 points_to_focus=BASE_POINTS, verbose=False):

        # Setting the API key to use the OpenAI API
        openai.api_key = openai_key
        try:
            http_proxy = nonebot.get_driver().config.openai_http_proxy
        except:
            http_proxy = ""

        if http_proxy != "":
            openai.proxy = {'http': http_proxy, 'https': http_proxy}

        # Initializing prompts for the conversation
        self.init_prompt = READING_PROMT_V2.format(points_to_focus)

        self.summary_prompt = 'You are a researcher helper bot. Now you need to read the summaries of a research paper.'
        self.messages = []  # Initializing the conversation messages
        self.summary_msg = []  # Initializing the summary messages
        self.token_len = token_length  # Setting the token length to use
        self.keep_round = 2  # Rounds of previous dialogues to keep in conversation
        self.model = model  # Setting the GPT model to use
        self.verbose = verbose  # Flag to enable/disable verbose logging
        self.model = OpenAIModel(api_key=openai_key, model=model)

    def drop_conversation(self, msg):
        # This method is used to drop previous messages from the conversation and keep only recent ones
        if len(msg) >= (self.keep_round + 1) * 2 + 1:
            new_msg = [msg[0]]
            for i in range(3, len(msg)):
                new_msg.append(msg[i])
            return new_msg
        else:
            return msg

    def send_msg(self, msg):
        return self.model.send_msg(msg)

    def _chat(self, message):
        # This method is used to send a message and get a response from the OpenAI API

        # Adding the user message to the conversation messages
        self.messages.append({"role": "user", "content": message})
        # Sending the messages to the API and getting the response
        response = self.send_msg(self.messages)
        # Adding the system response to the conversation messages
        self.messages.append({"role": "system", "content": response})
        # Dropping previous conversation messages to keep the conversation history short
        self.messages = self.drop_conversation(self.messages)
        # Returning the system response
        return response

    def summarize(self, paper: Paper):
        # This method is used to summarize a given research paper

        # Adding the initial prompt to the conversation messages
        self.messages = [
            {"role": "system", "content": self.init_prompt},
        ]
        # Adding the summary prompt to the summary messages
        self.summary_msg = [{"role": "system", "content": self.summary_prompt}]

        # Reading and summarizing each part of the research paper
        for (page_idx, part_idx, text) in paper.iter_pages():
            print('page: {}, part: {}'.format(page_idx, part_idx))
            # Sending the text to the API and getting the response
            summary = self._chat('now I send you page {}, part {}：{}'.format(page_idx, part_idx, text))
            # Logging the summary if verbose logging is enabled
            if self.verbose:
                print(summary)
            # Adding the summary of the part to the summary messages
            self.summary_msg.append({"role": "user", "content": '{}'.format(summary)})

        # Adding a prompt for the user to summarize the whole paper to the summary messages
        self.summary_msg.append({"role": "user", "content": 'Now please make a Chinese summary of the whole paper.'})
        # Sending the summary messages to the API and getting the response
        result = self.send_msg(self.summary_msg)
        # Returning the summary of the whole paper
        return result

    def read_pdf_and_summarize(self, pdf_path):
        # This method is used to read a research paper from a PDF file and summarize it

        # Creating a PdfReader object to read the PDF file
        pdf_reader = PdfReader(pdf_path)
        paper = Paper(pdf_reader)
        # Summarizing the full text of the research paper and returning the summary
        print('reading pdf finished')
        summary = self.summarize(paper)
        return summary

    def get_summary_of_each_part(self):
        # This method is used to get the summary of each part of the research paper
        return self.summary_msg

    def question(self, question):
        # This method is used to ask a question after summarizing a paper

        # Adding the question to the summary messages
        self.summary_msg.append({"role": "user", "content": question})
        # Sending the summary messages to the API and getting the response
        response = self.send_msg(self.summary_msg)
        # Adding the system response to the summary messages
        self.summary_msg.append({"role": "system", "content": response})
        # Returning the system response
        return response



BASE_POINTS = """
1. Who are the authors?
2. What is the process of the proposed method?
3. What is the performance of the proposed method? Please note down its performance metrics.
4. What are the baseline models and their performances? Please note down these baseline methods.
5. What dataset did this paper use?
"""

READING_PROMPT = """
You are a researcher helper bot. You can help the user with research paper reading and summarizing. \n
Now I am going to send you a paper. You need to read it and summarize it for me part by part. \n
When you are reading, You need to focus on these key points:{}
"""

READING_PROMT_V2 = """
You are a researcher helper bot. You can help the user with research paper reading and summarizing. \n
Now I am going to send you a paper. You need to read it and summarize it for me part by part. \n
When you are reading, You need to focus on these key points:{},

And You need to generate a brief but informative title for this part.
Your return format:
- title: '...'
- summary: '...'
"""

SUMMARY_PROMPT = "You are a researcher helper bot. Now you need to read the summaries of a research paper.Please answer in Chinese."



import asyncio
import io
import os
import shutil
from datetime import timedelta
from pathlib import Path
import requests

import nonebot
from nonebot import on_command, on_notice
from nonebot.adapters.onebot.v11 import (Message, MessageSegment)
from nonebot.adapters.onebot.v11 import MessageEvent, Bot, GroupMessageEvent, GroupUploadNoticeEvent
from nonebot.params import CommandArg

from .ChatSession import file2embedding, get_ans
from .gpt_reader.pdf_reader import PaperReader, BASE_POINTS

try:
    api_key = nonebot.get_driver().config.openai_api_key
except:
    api_key = ""

data = {}

embedding_folder = Path() / "data" / "nonebot-plugin-chatpdf" / "embedding"
result_folder = Path() / "data" / "nonebot-plugin-chatpdf" / "result"


def delete_file(dir):
    if os.path.exists(dir):
        shutil.rmtree(dir)


delete_file(result_folder)
delete_file(embedding_folder)

Path(result_folder).mkdir(parents=True, exist_ok=True)
Path(embedding_folder).mkdir(parents=True, exist_ok=True)

# start_request = on_command("/start", block=True, priority=1)


# @start_request.handle()
# async def _(event: MessageEvent, msg: Message = CommandArg()):
#     if api_key == "":
#         await start_request.finish(MessageSegment.text("请先配置openai_api_key"), at_sender=True)

#     data[event.get_session_id()] = ""

#     delete_file(os.path.join(result_folder, event.get_session_id()))
#     delete_file(os.path.join(embedding_folder, event.get_session_id()))

#     await start_request.finish(MessageSegment.text("chatpdf开始记录，请发送/add ，后接文本内容添加文档。/stop命令停止"), at_sender=True)


# add_request = on_command("/add", block=True, priority=1)


# @add_request.handle()
# async def _(event: MessageEvent, msg: Message = CommandArg()):
#     if api_key == "":
#         await add_request.finish(MessageSegment.text("请先配置openai_api_key"))

#     content = msg.extract_plain_text() + "\n"
#     if content == "" or content is None:
#         await add_request.finish(MessageSegment.text("内容不能为空！"))

#     if event.get_session_id() not in data:
#         await add_request.finish(MessageSegment.text("请先使用/start命令开始！"))

#     data[event.get_session_id()] += content
#     await add_request.finish(MessageSegment.text("添加成功！"), at_sender=True)


# stop_request = on_command("/stop", block=True, priority=1)


# @stop_request.handle()
# async def _(event: MessageEvent, msg: Message = CommandArg()):
#     if api_key == "":
#         await add_request.finish(MessageSegment.text("请先配置openai_api_key"))

#     stop_request.send(MessageSegment.text("开始分析......"))
#     loop = asyncio.get_event_loop()
#     try:
#         await loop.run_in_executor(None, text2embedding, event.get_session_id(), data[event.get_session_id()])
#     except Exception as error:
#         data[event.get_session_id()] = ""
#         await stop_request.finish(str(error))

#     data[event.get_session_id()] = ""
#     await stop_request.finish(MessageSegment.text("文章分析完成！"), at_sender=True)

pdf_request = on_command("/start", block=True, priority=1)


@pdf_request.handle()
async def _(event: GroupMessageEvent):
    pdf_file_request = on_notice(temp=True, expire_time=timedelta(minutes=1))
    id = event.get_session_id()

    data[id] = ""

    @pdf_file_request.handle()
    async def _(event: GroupUploadNoticeEvent):
        await pdf_request.send(MessageSegment.text("分析中，请稍等......"), at_sender=True)
        response = requests.get(event.file.url)

        with io.BytesIO(response.content) as f:
            session = PaperReader(api_key, points_to_focus=BASE_POINTS)
            data[id] = session
            summary = session.read_pdf_and_summarize(f)

        await pdf_request.finish(MessageSegment.text(summary + "现在，你可以向我提问有关于改文章的问题了"), at_sender=True)

    await pdf_request.finish(MessageSegment.text("请上传需要分析的pdf文件"), at_sender=True)


pdf_chat_request = on_command("/chat_pdf", block=True, priority=1)


@pdf_chat_request.handle()
async def _(bot: Bot, event: GroupMessageEvent, msg: Message = CommandArg()):
    question = msg.extract_plain_text()
    await pdf_request.finish(MessageSegment.text(data[event.get_session_id()].question(question)), at_sender=True)


txt_chat_request = on_command("/chat_txt", block=True, priority=1)


@txt_chat_request.handle()
async def _(event: MessageEvent, msg: Message = CommandArg()):
    question = msg.extract_plain_text()
    loop = asyncio.get_event_loop()

    try:
        ans = await get_ans(event.get_session_id(), question)

    except Exception as error:
        await txt_chat_request.finish(str(error))
    await txt_chat_request.finish(MessageSegment.text(ans), at_sender=True)


delete_request = on_command("/delete_all", block=True, priority=1)


@delete_request.handle()
async def _():
    delete_file(result_folder)
    delete_file(embedding_folder)
    data = {}
    await delete_request.finish(MessageSegment.text("全部删除文件缓存成功！"), at_sender=True)


delete_user_request = on_command("/delete_my", block=True, priority=1)


@delete_user_request.handle()
async def _(event: MessageEvent):
    delete_file(os.path.join(result_folder, event.get_session_id()))
    delete_file(os.path.join(embedding_folder, event.get_session_id()))
    data[event.get_session_id()] = ""
    await delete_user_request.finish(MessageSegment.text("成功删除你在该群的缓存文件！"), at_sender=True)


txt_request = on_command("/txt", block=True, priority=1)


@txt_request.handle()
async def test_(bot: Bot, msg: Message = CommandArg()):
    file_request = on_notice(temp=True, expire_time=timedelta(minutes=1))

    @file_request.handle()
    async def _(event: GroupUploadNoticeEvent):
        # logger.info(event.file.url)
        await file_request.send(MessageSegment.text("分析中，请稍等......"), at_sender=True)
        loop = asyncio.get_event_loop()
        try:
            await loop.run_in_executor(None, file2embedding, event.get_session_id(), event.file.url)
        except Exception as error:
            data[event.get_session_id()] = ""
            await file_request.finish(str(error))

        data[event.get_session_id()] = ""
        await file_request.finish(MessageSegment.text("文章分析完成！"), at_sender=True)

    await txt_request.finish(MessageSegment.text("请上传需要分析的txt文件"), at_sender=True)



import hashlib
import json
import os.path
import time
import urllib
from pathlib import Path

import nonebot
import numpy as np
import openai

try:
    api_key = nonebot.get_driver().config.openai_api_key
except:
    api_key = ""

try:
    http_proxy = nonebot.get_driver().config.openai_http_proxy
except:
    http_proxy = ""

try:
    time_limit = nonebot.get_driver().config.openai_api_limit
except:
    time_limit = "True"


if http_proxy != "":
    openai.proxy = {'http': http_proxy, 'https': http_proxy}

openai.api_key = api_key

COMPLETIONS_MODEL = "gpt-3.5-turbo"
EMBEDDING_MODEL = "text-embedding-ada-002"
CONTEXT_TOKEN_LIMIT = 3500


def get_embedding(text: str, event_id: str, model: str = EMBEDDING_MODEL):
    folder = Path() / "data" / "nonebot-plugin-chatpdf" / "embedding" / event_id
    Path(folder).mkdir(parents=True, exist_ok=True)
    tmpfile = os.path.join(folder, hashlib.md5(text.encode('utf-8')).hexdigest() + ".json")
    if os.path.isfile(tmpfile):
        with open(tmpfile, 'r', encoding='UTF-8') as f:
            return json.load(f)
    
    if time_limit == "True" :
        time.sleep(0.1)
    
    result = openai.Embedding.create(
        model=model,
        input=text
    )

    with open(tmpfile, 'w', encoding='utf-8') as handle2:
        json.dump(result["data"][0]["embedding"], handle2, ensure_ascii=False, indent=4)

    return result["data"][0]["embedding"]


def file2embedding(event_id: str, url):
    embeddings = []
    sources = []

    folder = Path() / "data" / "nonebot-plugin-chatpdf" / "result" / event_id
    Path(folder).mkdir(parents=True, exist_ok=True)

    with urllib.request.urlopen(url) as response:
        content = response.read().decode()
        # print(content)
    for source in content.split('\n'):
        if source.strip() == '':
            continue
        embeddings.append(get_embedding(source, event_id))
        sources.append(source)

    tmppath = os.path.join(folder, 'result.json')

    with open(tmppath, 'w', encoding='utf-8') as handle2:
        json.dump({"sources": sources, "embeddings": embeddings}, handle2, ensure_ascii=False, indent=4)


def text2embedding(event_id: str, contents=""):
    embeddings = []
    sources = []
    content = contents

    folder = Path() / "data" / "nonebot-plugin-chatpdf" / "result" / event_id
    Path(folder).mkdir(parents=True, exist_ok=True)
    for source in content.split('\n'):
        if source.strip() == '':
            continue
        embeddings.append(get_embedding(source, event_id))
        sources.append(source)

    tmppath = os.path.join(folder, 'result.json')

    with open(tmppath, 'w', encoding='utf-8') as handle2:
        json.dump({"sources": sources, "embeddings": embeddings}, handle2, ensure_ascii=False, indent=4)


def vector_similarity(x, y):
    """
    Returns the similarity between two vectors.
    
    Because OpenAI Embeddings are normalized to length 1, the cosine similarity is the same as the dot product.
    """
    return np.dot(np.array(x), np.array(y))


def order_document_sections_by_query_similarity(query: str, embeddings, event_id):
    """
    Find the query embedding for the supplied query, and compare it against all of the pre-calculated document embeddings
    to find the most relevant sections. 
    
    Return the list of document sections, sorted by relevance in descending order.
    """
    query_embedding = get_embedding(query, event_id)

    document_similarities = sorted([
        (vector_similarity(query_embedding, doc_embedding), doc_index) for doc_index, doc_embedding in
        enumerate(embeddings)
    ], reverse=True, key=lambda x: x[0])

    return document_similarities


async def ask(question: str, embeddings, sources, event_id):
    ordered_candidates = order_document_sections_by_query_similarity(question, embeddings, event_id)
    ctx = ""
    for candi in ordered_candidates:
        next = ctx + " " + sources[candi[1]]
        if len(next) > CONTEXT_TOKEN_LIMIT:
            break
        ctx = next
    if len(ctx) == 0:
        return ""

    prompt = "".join([
        u"基于上下文回答以下问题\n\n"
        u"上下文：" + ctx + u"\n\n"
                        u"问题：" + question + u"\n\n"
                                            u"答案："
    ])

    completion = await openai.ChatCompletion.acreate(model="gpt-3.5-turbo",
                                                     messages=[{"role": "user", "content": prompt}])
    return [prompt, completion.choices[0].message.content]


async def get_ans(event_id: str, question: str):
    folder = Path() / "data" / "nonebot-plugin-chatpdf" / "result" / event_id

    tmppath = os.path.join(folder, 'result.json')
    if os.path.isfile(tmppath):
        with open(tmppath, 'r', encoding='UTF-8') as f:
            obj = json.load(f)
            [prompt, answer] = await ask(question, obj["embeddings"], obj["sources"], event_id)

        return answer

    return "还未生成embeding！"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_chatpdf', 'nonebot_plugin_chatpdf.gpt_reader']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.4,<4.0.0',
 'nonebot-adapter-onebot>=2.2.1,<3.0.0',
 'nonebot2>=2.0.0rc3,<3.0.0',
 'numpy>=1.24.2,<2.0.0',
 'openai>=0.27.1,<0.28.0',
 'pypdf2>=3.0.1,<4.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-chatpdf',
    'version': '0.3.0',
    'description': 'A nonebot plugin for chatpdf',
    'long_description': '<div align="center">\n  <a href="https://v2.nonebot.dev/store"><img src="https://github.com/A-kirami/nonebot-plugin-template/blob/resources/nbp_logo.png" width="180" height="180" alt="NoneBotPluginLogo"></a>\n  <br>\n  <p><img src="https://github.com/A-kirami/nonebot-plugin-template/blob/resources/NoneBotPlugin.svg" width="240" alt="NoneBotPluginText"></p>\n</div>\n\n<div align="center">\n\n# nonebot-plugin-chatpdf\n\n</div>\n\n# 介绍\n\n- 本插件灵感来源于最近很火的[chatpdf](https://www.chatpdf.com)。\n- 将需要分析的论文/阅读材料上传到群文件，或者分次发送给机器人，机器人可以对其进行存储分析，然后你可以向其提问有关文章内容、文章概要、对于文章的思考等问题\n- 本插件参考和使用了项目[ChatGPT-Paper-Reader](https://github.com/talkingwallace/ChatGPT-Paper-Reader)\n  和[How to Code a Project like ChatPDF?](https://postor.medium.com/how-to-code-a-project-like-chatpdf-e40441cb4168)中的代码\n\n# 效果\n\n![Alt](./img/img1.jpg)\n![Alt](./img/img2.jpg)\n\n# 安装\n\n* 手动安装\n  ```\n  git clone https://github.com/Alpaca4610/nonebot_plugin_chatpdf.git\n  ```\n\n  下载完成后在bot项目的pyproject.toml文件手动添加插件：\n\n  ```\n  plugin_dirs = ["xxxxxx","xxxxxx",......,"下载完成的插件路径/nonebot-plugin-chatpdf"]\n  ```\n* 使用 pip\n  ```\n  pip install nonebot-plugin-chatgpt-chatpdf\n  ```\n\n# 配置文件\n\n在Bot根目录下的.env文件中追加如下内容：\n\n```\nOPENAI_API_KEY = key\n```\n\n可选内容：\n\n```\nOPENAI_HTTP_PROXY = "http://127.0.0.1:8001"    # 设置代理解决OPENAI的网络问题\nOPENAI_API_LIMIT = True     # 降低生成embedding时向OpenAI发送请求的频率，提高embedding生成的成功率。默认开启，改为False关闭。免费版API建议开启。\n```\n\n# 使用方法\n\n如果设置了nonebot全局触发前缀，需要在下面的命令前加上设置的前缀。\n\n### 使用方式1：上传需要分析的pdf文件到群文件中\n\n- /start (使用该命令以上传pdf文件的方式启动chatpdf文章分析功能)\n- 在一分钟内，上传需要分析的pdf文件到群文件中，分析完成后会返回成功信息\n- /chat_pdf (文章分析完成后，使用该命令后面接需要提问的关于文章的问题，机器人会给出答案)\n- /delete_all (删除所有缓存文件)\n- /delete_my （删除用户在本群的缓存文件）\n\n### 使用方式2：上传需要分析的txt文件到群文件中\n\n- /txt (使用该命令以上传文件的方式启动chatpdf文章分析功能)\n- 在一分钟内，上传需要分析的txt文件到群文件中，机器人会对其进行分析并使用OpenAI的API生成embedding文件，分析完成后会返回成功信息\n- /chat_txt (文章分析完成后，使用该命令后面接需要提问的关于文章的问题，机器人会给出答案)\n- /delete_all (删除所有缓存文件)\n- /delete_my （删除用户在本群的缓存文件）\n\n# 注意事项\n\n- 使用txt分析的过程中会在机器人的data文件夹下产生embedding缓存文件，注意缓存占用\n- 每次调用/start命令时，都会清除调用者以前的分析缓存\n- 插件加载时会删除所有用户的embedding缓存文件\n',
    'author': 'Alpaca',
    'author_email': 'alpaca@bupt.edu.cn',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List
import openai


class ModelInterface(object):

    def __init__(self) -> None:
        pass

    def send_msg(self, *args):
        pass


class OpenAIModel(object):

    def __init__(self, api_key, model='gpt-3.5-turbo', temperature=0.2) -> None:
        openai.api_key = api_key
        self.model = model
        self.temperature = temperature

    def send_msg(self, msg: List[dict], return_raw_text=True):
        
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=msg,
            temperature=self.temperature
        )

        if return_raw_text:
            return response["choices"][0]["message"]["content"]
        else:
            return response
